#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment procedures

# (c) Issabella (tutorial) / Dinasset (filter) / 2021
#
#   History:
#
#   v1.0: 2021-07-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import random
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
FixWH=3000

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()
        
# here the single GEGL functions

def gegl_c2g(image, drawable, uradius, usamples, uiter):
    radius=int(uradius)      #dft 300   #ui_range (2, 1000)
    samples=int(usamples)    #dft 4     #ui_range (3, 17)
    iterations=int(uiter)    #dft 10    #ui_range (1, 30)
    enhance_shadows=False
    mystring="gegl:c2g radius=%d samples=%d iterations=%d enhance-shadows=%d"
    gegl_graph_string=mystring % (radius, samples, iterations, enhance_shadows)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_edge(image, drawable, ualgorithm): 
    algorithm="sobel" # 0=sobel,1=prewitt,2=gradient,3=roberts,4=differential,5=laplace
    amount=2.0        # value_range  (1.00 to 10.0)
    border=1          # 0=none,1=clamp,2=loop,3=black,4=white
    mystring="gegl:edge algorithm=%s amount=%f border=%d"
    gegl_graph_string=mystring % (algorithm, amount, border)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_median_blur (image, drawable, uradius):
    radius=float(uradius)
    percentile=50.0
    alpha_percentile=50.0
    gegl_graph_string="gegl:median-blur radius=%f percentile=%f alpha-percentile=%f" % (radius, percentile, alpha_percentile)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_gaussian_blur (image, drawable, ux, uy):    # gaussian blur  
    std_dev_x=float(ux)    #1.50     #user_range   (0.24, 100.0)
    std_dev_y=float(uy)    #1.50     #user_range   (0.24, 100.0)
    mystring="gegl:gaussian-blur std-dev-x=%f std-dev-y=%f"
    gegl_graph_string=mystring % (std_dev_x, std_dev_y)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_noise_cie_lch (image, drawable, uhold, ulightd, uchromad, uhued, useed):
    holdness=int(uhold)
    lightness_distance=float(ulightd)
    chroma_distance=float(uchromad)
    hue_distance=float(uhued)
    seed=int(useed)
    mystring="gegl:noise-cie-lch holdness=%d lightness-distance=%f chroma-distance=%f hue-distance=%f seed=%d"
    gegl_graph_string=mystring % (holdness, lightness_distance, chroma_distance, hue_distance, seed)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_imagegradient (image, drawable, umode):
    output_mode=umode
    gegl_graph_string="gegl:image-gradient output-mode=%s" % (output_mode)
    gegl_graph(image, drawable,gegl_graph_string)
 
    
############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    optiWidthHeight = FixWH       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    # adjust mean values
    adjustmeanv (newImage, newBG)
    # remove low values, remove high values  
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                   1.00, 0.03, 0.97, False)
    newBG.opacity = 100.0
    
    return(newImage)

#---------------------------------------------------------------------------
# adjust values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer):
    # new version modified to take care of the existing mean in determining the theretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    # consider 3 classes
    if meanl < 85: thmean = (128+meanl)/2  #100
    elif meanl > 170: thmean = (127+meanl)/2  #155
    else: thmean = 127.5    # regular
    minmean = thmean-16
    maxmean = thmean+16
    
    modified=False
    if meanl < minmean:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        loop=0
        while meanl < minmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.00,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        loop=0
        while meanl > maxmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.00,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))
    
    return(inLayer)
     
#
# ================ *
#   M  A  I  N     *
# ---------------- *
   
def JvidEffectHR2 (inImage, inLayer, widthheight, 
                 doFlatten):
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # display messages using error console
    
    newImage = createOutimage (inImage, inLayer)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width = Preprocessed.width
    height = Preprocessed.height
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # - Median Blur
    MedianBlur = Preprocessed.copy()
    newImage.add_layer (MedianBlur, 0)
    MedianBlur.name = "Median Blurred 5.0"
    RADIUS=5.0
    gegl_median_blur (newImage, MedianBlur, RADIUS)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # GEGL Image Gradient
    Gradient = MedianBlur.copy()
    newImage.add_layer (Gradient, 0)
    Gradient.mode = LAYER_MODE_MULTIPLY  #GRAIN_MERGE  15.07.21
    GType="magnitude"
    Gradient.name = "Gradient "+GType
    gegl_imagegradient (newImage, Gradient, GType)
     
    pdb.gimp_drawable_invert (Gradient, False)
    pdb.gimp_drawable_threshold (Gradient, HISTOGRAM_VALUE, 0.75,1.0)  #15.07.21
     
    #pdb.gimp_display_new (newImage)
    #return
     
    # prepare first base  
    Base1 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base1, 0)
    Base1.name = "Base for Hard Sketh (First Intermediate Outcome)"
    Base1.mode = LAYER_MODE_NORMAL
    
    # 1.st effect: Hard Sketch
    pdb.plug_in_gmic_qt(newImage, Base1, 1,2 ,       # just try no blur
   "-v - -fx_hardsketchbw 900,100,3,0.1,20,0,4,0,50,50 ") # just try
    HardSketch_Layer = newImage.layers[0]
    HardSketch_Layer.mode = LAYER_MODE_BURN #GRAIN_MERGE #MULTIPLY #LINEAR_BURN  15.07.21
    HardSketch_Layer.opacity = 44.0   #100.0  15.07.21
    HardSketch_Layer.name = "[GMIC] -fx_hardsketchbw 1500,50,1,0.1,20,0,4,0,50,50 "
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # prepare second base  
    Base2 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base2, 0)
    Base2.name = "Hard Sketched (Second Intermediate Outcome)"
    Base2.mode = LAYER_MODE_NORMAL
    
    # add a posterized source
    Poster = Preprocessed.copy()
    newImage.add_layer (Poster, 0)
    Poster.mode = LAYER_MODE_LCH_COLOR  # changed from MULTIPLY (15.07.21)
    Poster.opacity = 100.0
    pdb.gimp_drawable_posterize (Poster, 32)
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # prepare third base  
    Base3 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base3, 0)
    Base3.name = "Base for Sketch (Third Intermediate Outcome)"
    Base3.mode = LAYER_MODE_NORMAL
    
    # add some noise
    HOLD=2; LIGHTD=45.0; CHROMAD=45.0; HUED=7.5; SEED=0
    gegl_noise_cie_lch (newImage, Base1, HOLD, LIGHTD, CHROMAD, HUED, SEED)
    # check luma
    adjustmeanv (newImage, Base3)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # 2.nd effect: Sketch
    pdb.plug_in_gmic_qt(newImage, Base1, 1,2 ,
    #[G'MIC] Sketch: fx_sketchbw 3,45,180,30,1.75,0.02,0.5,0.75,0.1,0.7,3,6,0,1,4,0,50,50
    #[G'MIC] Sketch: fx_sketchbw 3,45,180,90,1.75,0.02,0.5,0.75,0.1,0.7,3,10,0,1,4,0,50,50
    # samj_fx_sketchbw_modifie 3,45,180,30,1.75,0.02,0.5,0.75,0.1,0.7,3,6,0,1,4,0,1234,0,2,100,0.2,1,4,1,0
    "-v - samj_fx_sketchbw_modifie "+
    "3,45,180,30,1.75,0.02,0.5,0.75,0.1,0.7,3,6,0,1,4,0,1234,0,2,100,0.2,1,4,1,0 ")
    Sketch_Layer = newImage.layers[0]
    Sketch_Layer.mode = LAYER_MODE_GRAIN_MERGE
    Sketch_Layer.opacity = 100.0  
    Sketch_Layer.name = "[GMIC] samj_fx_sketchbw 3,45,180,90,..."
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # duplicate and set to Grain Merge  
    Sketch_Layer2 = Sketch_Layer.copy()
    newImage.add_layer (Sketch_Layer2, 0)
    Sketch_Layer2.mode = LAYER_MODE_SCREEN  # added instead of default GRAIN MERGE 15.07.21
    Sketch_Layer2.opacity = 50.0  
    #pdb.gimp_drawable_curves_spline (Sketch_Layer2, HISTOGRAM_VALUE,
    #                                 4, (0.0,0.0,1.0,0.5))
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # prepare second base  
    Base4 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base4, 0)
    Base4.name = "Base for Paint (Fourth Intermediate Outcome)"
    Base4.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
      
    # 3.rd effect: Vector Paint
    pdb.plug_in_gmic_qt(newImage, Base4, 1,2 ,
    #[G'MIC] VectorPainting: fx_vector_painting 9.8,0,50,50
    "-v - -fx_vector_painting 9.82,0,50,50 ")
    VPaint_Layer = newImage.layers[0]
    VPaint_Layer.mode = LAYER_MODE_GRAIN_MERGE
    VPaint_Layer.opacity = 25.0  
    VPaint_Layer.name = "[GMIC] -fx_vector_painting 9.8,0,50,50"
    
    #pdb.gimp_display_new (newImage)
    #return
      
    # prepare fifth base  
    Base5 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base5, 0)
    Base5.name = "Base for Adding some source (Fifth Intermediate Outcome)"
    Base5.mode = LAYER_MODE_NORMAL
     
    # equalize partially
    Base5E = Base5.copy()
    newImage.add_layer (Base5E, 0)
    pdb.gimp_drawable_equalize (Base5E, False)
    Base5E.opacity = 10.0
    Base5 = pdb.gimp_image_merge_down (newImage, Base5E, CLIP_TO_IMAGE)
    # add canvas
    pdb.plug_in_gmic_qt(newImage, Base5, 1,0 ,
    #[G'MIC] Canvas: fx_canvas 70,45,1956,1,70,135,400,0,50,50
    "-v - -fx_canvas 70,45,1956,1,70,135,400,0,50,50 ")
    # check luma
    adjustmeanv (newImage, Base5)
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # copy Preprocessed above
    PPonTop = Preprocessed.copy()
    newImage.add_layer (PPonTop, 0)
    PPonTop.name = "Bits of source colours (Preprocessed copy)"
    PPonTop.mode = LAYER_MODE_GRAIN_MERGE
    PPonTop.opacity = 75.0  #50.0 chg 15.07.21
    
    PPedges = PPonTop.copy()
    newImage.add_layer (PPedges, 0)
    PPedges.name = "Source edges"
    (meanPP,b,c,d,e,f)=pdb.gimp_drawable_histogram(PPonTop,HISTOGRAM_VALUE,0.0,1.0)
    PPedges.mode = LAYER_MODE_GRAIN_MERGE # changed from LINEAR_BURN 15.07.21
    PPedges.opacity = 35.0  # CHG FROM 55.0  15.07.21
    # edges (apply edge (sobel)
    ALGORITHM="sobel"
    gegl_edge (newImage, PPedges, ALGORITHM)
    pdb.plug_in_vinvert (newImage, PPedges)
    
    #pdb.gimp_display_new (newImage)
    #return
     
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "outcome"
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with JVID effect HR2 (Rough Sketch Graphic Look) "
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="issabella/dinasset"
menu="<Image>/Jvid/"
copyrightYear="2021"
desc="JVID Effect HR2 (Delicate Sketch Graphic Look)"

register(
    "JvidEffectHR2",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 6000, 100)),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    JvidEffectHR2,
    menu=menu
)

main()
